/* qos.c - quality of service routines */

#define qos_c
#ifdef QOS

#include <math.h>
#include <unistd.h>
#include <stdio.h>
#include "quipu/util.h"
#include "messages.h"
#include "demanifest.h"
#include "qos.h"

enum QosModes { QOS_NORMAL_UPDATE, QOS_BULK_UPDATE, QOS_OFF };

#define QOS_MIN_MESSAGE_TIME 2 
#define QOS_MIN_UPDATE_TIME 5
#define QOS_MAX_MESSAGE_TIME 180
#define QOS_BULK_EXTRA_INIT_TIME 7 
#define QOS_MIN_ABANDON_TIME 15
#define QOS_SEARCH_RETURN_MARGIN 5

static void writeSingleQosRec();
static void rewriteQosRec();
static void writeNewQosRec();
static void setNewQosRec();
static void setQos_dn();
static void setOldQosRec();
static void fputQosrec();
static void doInitialTimeStamp();
static void standardWaitMessage();
static void strip_ou();
static void formatWaitMessage();
static void doFinalTimeStamp();

static float reCalcAveTime();

static int checkBulkInitOffset();
static int checkForFastOKResponse();
static int checkForSlowResponse();
static int checkForFastAbandon();
static int checkQosFileSize();
static int calcElapsedTime();
static int lockQosFile();
static int unlockQosFile();
static int lockQosRec();
static int unlockQosRec();
static int readqosfile();
static int reReadRec();
static int buf2QosRec();
static int findqosrec();
static int tookAbout();

extern char qosfilename[];
extern int qosOn;

static FILE *qos_file;
static struct qosrec qosdata[QOSRECS];
static int noQosRecs;
static int cantAddRecs = FALSE;
static int qosMode;
static time_t startInitTime, endInitTime;

char waitMessage[BUFSIZ];
int qosAlarm;

void
initQos()
{
  if (qosOn == TRUE)
  {
    noQosRecs = readqosfile();
    if (noQosRecs == -1)
    {
        switchOffQos();
    }
    else
    {
      qosMode = QOS_NORMAL_UPDATE;
    }
  }
  else
  {
      qosMode = QOS_OFF;
  }
}

void
switchOffQos()
{
  qosOn = FALSE;
  qosMode = QOS_OFF;
}

/* An interface may want to do searches in asynchronous mode by sending 
batch searches and waiting for the results */
void
bulkQosUpdateOn()
{
    if(qosMode != QOS_OFF)
    {
        qosMode = QOS_BULK_UPDATE;
        startInitTime = time(0);
        endInitTime = 0;
    }
}

void 
setQosOffsetTime()
{
    if( qosMode != QOS_BULK_UPDATE )
        return;

    endInitTime = time(0);

#ifdef QOS_DEBUG
    (void)fprintf(stderr, "\n\nSearch Initialisation time %d secs\n",
                                          (int) (endInitTime - startInitTime));
#endif QOS_DEBUG
}
    
void
bulkQosUpdateOff()
{
    if(qosMode != QOS_OFF )
    {
        qosMode = QOS_NORMAL_UPDATE;
    }
}

/* finds record or creates a new one if can't find one.  
   formats approp message to display 
   to user; otherwise doesn't!
   Time-stamps record */
void
startQosTime(buf, qosent)
struct qosentry *qosent;
char *buf;
{
  int recno;
  char sysbuf[BUFSIZ];

  if ( qosMode == QOS_OFF )
  {
      return;
  }
  if( qosMode == QOS_BULK_UPDATE)
  {
      if( endInitTime != 0 )
      {

#ifdef QOS_DEBUG
          (void)fprintf(stderr,"Search Initiated Late Cannot do Qos %.30s\n",buf);
#endif QOS_DEBUG

          qosent->recnum = ERRQOSREC;
          return;
      }
  }
  strip_ou(qosent->matchString, buf);
  if( strlen(qosent->matchString) >= DNSIZE )  /* too long to add to table */
  {

#ifdef QOS_DEBUG
          (void)fprintf(stderr,"Stripped dn_string too large to add to table %.30s\n",
                                                          qosent->matchString);
#endif QOS_DEBUG
 
     formatWaitMessage(ERRQOSREC);
     qosent->recnum = ERRQOSREC;
     return;
  }

  findqosrec(qosent);

#ifdef QOS_DEBUG
  (void)fprintf(stderr,"record %d name %.30s\n", qosent->recnum, qosent->matchString);
#endif QOS_DEBUG

  if ((qosent->recnum != ERRQOSREC) && (qosent->recnum != NEWQOSREC))
    if (qosdata[qosent->recnum].noQueries > 0) /* only time initial query - once connection
                                esatablished, will probably go quite a bit
                                quicker */
    {
      formatWaitMessage(ERRQOSREC);
      return;
    }

  if( cantAddRecs == TRUE && qosent->recnum == NEWQOSREC)
  {
      formatWaitMessage(ERRQOSREC);

#ifdef QOS_DEBUG
      (void)fprintf(stderr, "Can't do QOS - can't add new record %30.s\n", 
                                                          qosent->matchString);
#endif QOS_DEBUG

      qosent->recnum = ERRQOSREC;
      return;
  }
  formatWaitMessage(qosent->recnum);
  doInitialTimeStamp(qosent);  
}

void
updateStats(opok, qosent)
int opok;
struct qosentry *qosent;
{
    if ((qosMode == QOS_OFF) || (qosent->recnum == ERRQOSREC))
        return;
    if ((qosMode == QOS_BULK_UPDATE) && (opok == QOS_ABANDON))
        return; /* don't record abandons in bulk mode - user might have 
                   results he/she wanted */
    if (qosdata[qosent->recnum].noQueries > 0)
	return; /* only record first query to a domain */
    doFinalTimeStamp(qosent);
    if (qosMode != QOS_BULK_UPDATE)
    {
      if ( checkForSlowResponse(qosent) ||
           checkForFastAbandon(opok, qosent) )
          return;

      if (qosent->recnum == NEWQOSREC)
       if (checkForFastOKResponse(opok, qosent))
         return;
    }

    writeSingleQosRec(opok, qosent);
    qosent->recnum = ERRQOSREC; /* in case it's rewritten without being *
                                 * initialised again                    */
}

/* check for OK fast response, as if we get one we won't bother to 
   update the QOS database */
static int
checkForFastOKResponse(opok, qosent) 
int opok;
struct qosentry *qosent;
{
    if (opok == QOS_FAILURE)
	return 0;
    if( calcElapsedTime(qosent) <= QOS_MIN_UPDATE_TIME )
    {
#ifdef QOS_DEBUG
          (void)fprintf(stderr,"Fast response for %.30s\n",
                                                          qosent->matchString);
#endif QOS_DEBUG
        return 1;
    }
    return 0;
}

static int
checkForSlowResponse(qosent) /* if response is too long may be */
struct qosentry *qosent;     /* an unrelated problem           */ 
{
    if( calcElapsedTime(qosent) > QOS_MAX_MESSAGE_TIME )
    {  
 
#ifdef QOS_DEBUG
          (void)fprintf(stderr,"Not writing record for %.30s as too long\n",
                                                          qosent->matchString);
#endif QOS_DEBUG
 
        return 1;
    }
    return 0;
}

static int 
checkForFastAbandon(opok, qosent) /* if operation is abandoned quickly then */
int opok;                         /* it's probably a user mistake           */
struct qosentry *qosent;
{
    if( opok == QOS_ABANDON && calcElapsedTime(qosent) < QOS_MIN_ABANDON_TIME)
    {

#ifdef QOS_DEBUG
        (void)fprintf(stderr,"Not writing record for %.30s under min abandon time\n",
                                                          qosent->matchString);
#endif QOS_DEBUG

        return 1; 
    }
    return 0;
}

static void 
writeSingleQosRec(opok, qosent)
int opok;
struct qosentry *qosent;
{
    if(qosent->recnum != NEWQOSREC)  /* existing record */
    {

#ifdef QOS_DEBUG
        (void)fprintf(stderr,"Trying to update existing QOS record %.30s\n",
                                                           qosent->matchString);
#endif QOS_DEBUG

        if( lockQosRec(qosent->recnum) == -1 )
            return;
        rewriteQosRec(opok, qosent);
        unlockQosRec(qosent->recnum);
    }
    else  /* new record */
    {
        if( noQosRecs < QOSRECS )
        {

#ifdef QOS_DEBUG
            (void)fprintf(stderr,"Trying to write new QOS record %.30s\n",
                                                           qosent->matchString);
#endif QOS_DEBUG

            if( checkQosFileSize() == -1 || lockQosFile() == -1 )
                return;

            writeNewQosRec(opok, qosent);
            unlockQosFile();
        }
        else
        {
            cantAddRecs = TRUE;

#ifdef QOS_DEBUG
            (void)fprintf(stderr,"QOS file full - can't add records\n");
#endif QOS_DEBUG
        
        }
    }
    qosdata[qosent->recnum].noQueries++;
}

static void 
rewriteQosRec(opok, qosent) /* rewrite existing record */
int opok;
struct qosentry *qosent;
{
    if(reReadRec(qosent->recnum)==-1) 
        return;           /* file corrupted qos switched off */
    setOldQosRec(opok, qosent);
    fputQosrec(qosent->recnum);
}

static void
writeNewQosRec(opok, qosent) /* write new record */
int opok;
struct qosentry *qosent;
{ 
    setNewQosRec(opok, qosent);
    qosent->recnum = noQosRecs++;
    fputQosrec(qosent->recnum); 
}

static int
checkQosFileSize()
{
    struct stat f_stat;

    if( fstat(fileno(qos_file), &f_stat) == -1 )
        return -1;

    if((f_stat.st_size / QOSRECSIZE) != noQosRecs)
    {

#ifdef QOS_DEBUG
        (void)fprintf(stderr, "File Has changed size - Cannot add records\n");
#endif QOS_DEBUG

        cantAddRecs = TRUE;
        return -1;
    }
    return 0;
}

static int
calcElapsedTime(qosent)
struct qosentry *qosent;
{
    return (int) (qosent->finalTime - qosent->startTime);
}

static int
checkBulkInitOffset(qosent) /* Bulk initialisation means times are invalid */
struct qosentry *qosent;    /* before end of initialisation                */
{
    if(qosMode != QOS_BULK_UPDATE)
        return calcElapsedTime(qosent);

/* I don't think the times are accurate when derived by bulk update
   let's return -1 for now and see if the figures look better (PB 5/11/93) */
    return -1;
#ifdef FOO
    if( qosent->finalTime <= endInitTime+QOS_BULK_EXTRA_INIT_TIME )
    {

#ifdef QOS_DEBUG
        (void)fprintf(stderr, "Time faster than end of initialisation\n");
#endif QOS_DEBUG

        return -1;
    }
    return calcElapsedTime(qosent);
#endif FOO
}
    
static void
setNewQosRec(opok, qosent)
int opok;
struct qosentry *qosent;
{
    /* new Entries are always added at the end of the list */

    setQos_dn(qosdata[noQosRecs].dn, qosent->matchString);

    qosdata[noQosRecs].timeStamp = qosent->startTime;
    qosdata[noQosRecs].cnt = 1; 

    switch(opok)
    {
        case QOS_SUCCESS:
            qosdata[noQosRecs].avetime = checkBulkInitOffset(qosent);
            qosdata[noQosRecs].success = QOS_SUCCESS_CHAR;
            break;

        case QOS_ABANDON:
            qosdata[noQosRecs].avetime = -1;/* record failure with no average */
            qosdata[noQosRecs].success = QOS_ABANDON_CHAR;
            break;

        default: /* treat default as failure */
            qosdata[noQosRecs].avetime = -1;/* record failure with no average */
            qosdata[noQosRecs].success = QOS_FAILURE_CHAR;
    }
}

static void
setQos_dn(dn, matchString)
char *dn;
char *matchString;
{
    int len;

    len = strlen(matchString);
    memcpy(dn, matchString, len);
    memset(dn + len, ' ', DNSIZE - len - 1);
    dn[DNSIZE] = '\0';
}

static void
setOldQosRec(opok, qosent)
int opok;
struct qosentry *qosent;
{
    switch(opok)
    {
        case QOS_SUCCESS:
        {
            int etime;

            qosdata[qosent->recnum].timeStamp = qosent->startTime;      

            if(qosdata[qosent->recnum].success != QOS_SUCCESS_CHAR)
            {  /* last record not a success */
                qosdata[qosent->recnum].cnt = 1;
            }
            else /* last record a success */
            {
                ++qosdata[qosent->recnum].cnt;
            }
            if((etime = checkBulkInitOffset(qosent))!= -1)
            {   /* only if valid time */
                if(qosdata[qosent->recnum].avetime == -1) /* no time recorded */
                {                                         /* so far           */
                    qosdata[qosent->recnum].avetime = etime;
                }
                else  /* valid elapsed time */
                {
                    qosdata[qosent->recnum].avetime = 
                    reCalcAveTime(qosdata[qosent->recnum].avetime, etime);
                }
            }
            qosdata[qosent->recnum].success = QOS_SUCCESS_CHAR;
            break;
        }
        case QOS_ABANDON:
        {
            if(qosdata[qosent->recnum].success == QOS_SUCCESS_CHAR)
            {   /* last record a success */
                qosdata[qosent->recnum].cnt = 1;
                qosdata[qosent->recnum].timeStamp = qosent->startTime;
            }
            else /* last record was a failure or abandon */
            {
            ++qosdata[qosent->recnum].cnt;
            }
            qosdata[qosent->recnum].success = QOS_ABANDON_CHAR;
            break;
        }
        default: /* failure */
        {
            if(qosdata[qosent->recnum].success == QOS_SUCCESS_CHAR)
            {   /* last record a success */
                qosdata[qosent->recnum].cnt = 1;
                qosdata[qosent->recnum].timeStamp = qosent->startTime;
            }
            else /* last record was a failure or abandon */
            {
            ++qosdata[qosent->recnum].cnt;
            }
            qosdata[qosent->recnum].success = QOS_FAILURE_CHAR;
        }
    } 
}

static float
reCalcAveTime(oldavetime, etime)
float oldavetime;
int etime;
{
    /* give quarter weight to this timing cf moving average */
    return (float) ((oldavetime * 3 + (float) etime) / 4) ;      
}

static void
fputQosrec(recno)
int recno;
{
    char outbuf[QOSRECSIZE+1];

    (void)sprintf(outbuf, "%-*s&%8.2f &%10ld%c &%4d\n", 
         DNSIZE,qosdata[recno].dn,
         qosdata[recno].avetime, qosdata[recno].timeStamp,
         qosdata[recno].success, qosdata[recno].cnt);

    fseek(qos_file, qosrecOffset(recno) , SEEK_SET); 
    fputs(outbuf, qos_file);
    fflush(qos_file);
}

static void 
doInitialTimeStamp(qosent)
struct qosentry *qosent;
{
    qosent->startTime = time(0);
}

static void
doFinalTimeStamp(qosent)
struct qosentry *qosent;
{
    qosent->finalTime = time(0);
}

static int
lockQosFile()
{
  if (fseek(qos_file, 0, SEEK_SET) == -1)
    return -1;
  return (lockf(fileno(qos_file), F_TLOCK, 0));
}

static int
unlockQosFile()
{
  if (fseek(qos_file, 0, SEEK_SET) == -1)
    return -1;
  return (lockf(fileno(qos_file), F_ULOCK, 0));
}

static int
lockQosRec(recno)
int recno;
{
  if (fseek(qos_file, qosrecOffset(recno) , SEEK_SET) == -1)
    return -1;

  return lockf(fileno(qos_file),  F_TLOCK, QOSRECSIZE);
}

static int 
unlockQosRec(recno)
int recno;
{
  fseek(qos_file, qosrecOffset(recno) , SEEK_SET);
  return lockf(fileno(qos_file), F_ULOCK, QOSRECSIZE);
}

static int
readqosfile()
{
  char inbuf[QOSRECSIZE];
  int i;

  if ((qos_file = fopen(qosfilename, "r+")) == (FILE *) NULL)
  {
    mode_t saveMask;

    saveMask = umask(RWRWRW_MASK);
    (void)fprintf(stderr,"Unable to find file QOS file '%s' ... trying to create!!\n",
                                                                 qosfilename);
    if((qos_file = fopen(qosfilename, "w+")) == (FILE *) NULL)
    {
        (void)fprintf(stderr,"Unable to create QOS file '%s'\n",qosfilename);
        (void)umask(saveMask);
        return -1; /* no can open */
    }
    else
    {
        (void)umask(saveMask);
        return 0; /* new file created with zero records */
    }
  }
  for( i = 0; i<= QOSRECS; ++i)
  {
      if(fgets(inbuf, QOSRECSIZE+1, qos_file) == NULLCP)
      {
          break;
      }
      if( buf2QosRec(inbuf, i) == -1 )
      {
          return -1;      /* file corrupted in some way */
      }
  }
  return i; /* no of recs */
}

static int
reReadRec(recno)
{
    char inbuf[QOSRECSIZE];

    fseek(qos_file, qosrecOffset(recno) , SEEK_SET);
    if (fgets(inbuf, QOSRECSIZE+1, qos_file) != NULLCP)
    if(buf2QosRec(inbuf, recno) == -1)
    {
        switchOffQos();
        return -1;
    }
    return 0;
}

static int
buf2QosRec(buf, recno)
char * buf;
int recno;
{
    if( sscanf(buf, "%[^&] & %f & %ld%c & %d", qosdata[recno].dn,
      &qosdata[recno].avetime, &qosdata[recno].timeStamp, 
      &qosdata[recno].success, &qosdata[recno].cnt) != 5 )
    {

#ifdef QOS_DEBUG
        (void)fprintf(stderr,"QOS file corrupted Switching QOS off\n");
#endif QOS_DEBUG

        return -1;
    }
    qosdata[recno].dn[DNSIZE] = '\0';
    qosdata[recno].noQueries = 0;
    return 0;
}

static int
findqosrec(qosent)
struct qosentry *qosent;
{
int i;
    
  (void)strcat(qosent->matchString," ");
  for(i = 0; i < noQosRecs; i++)
  {
    if (!(strncmp(qosdata[i].dn, qosent->matchString, 
                                               strlen(qosent->matchString))))
    {
      /* record matched*/
      qosent->recnum = i;
      return ;
    }
  }
  qosent->recnum = NEWQOSREC;
}


static
getMessageInfo(returnBuf, recno)
char *returnBuf;
int recno;
{
    char tmp[LINESIZE];
    char stampTime[LINESIZE];
 
    if( recno == NEWQOSREC )
    { 
        (void)strcpy(returnBuf, "\nNew Record - no information\n");
        return;
    }
    (void)strcpy(stampTime, ctime(&(qosdata[recno].timeStamp)));
    stampTime[strlen(stampTime)-1] = '\0';
    
    if(qosdata[recno].success == QOS_FAILURE_CHAR || 
       qosdata[recno].success == QOS_ABANDON_CHAR)
    {
        if( qosdata[recno].success == QOS_FAILURE_CHAR )
        {
            (void)sprintf(tmp, "\nLast enquiry was a failure, ");
        }
        else
        {
            (void)sprintf(tmp, "\nLast enquiry was abandoned, ");
        }
        (void)sprintf(tmp,"%s %d %s %s%s", tmp, qosdata[recno].cnt, 
        "enquiries to server\nhave failed/abandoned since",stampTime,
        "\nBefore failure/abandon");
    }
    else
    {   (void)sprintf(tmp, 
         "\nLast enquiry %s was successful\nLast", stampTime);
    }
    if( qosdata[recno].avetime != -1 )
    {
        (void)sprintf(returnBuf, "%s average enquiry time was %0.2f secs\n", tmp,
                                       qosdata[recno].avetime);
    }
    else
    {
        (void)sprintf(returnBuf, "%s average enquiry time not recorded\n", tmp);
    }
}

static void
standardWaitMessage()
{
    qosAlarm = 15;
    (void)sprintf(waitMessage, "%s%s",OPSOMETIME,PLEASEWAIT);
}

static void  
formatWaitMessage(recno)
int recno;
{
    if( qosMode != QOS_NORMAL_UPDATE )
    {
        return;
    }
    if( recno == NEWQOSREC || recno == ERRQOSREC )
    {
        standardWaitMessage();
    }
    else if( (qosdata[recno].success == QOS_ABANDON_CHAR || 
              qosdata[recno].success == QOS_FAILURE_CHAR) && 
              qosdata[recno].cnt > 4 )
    {
        char tmp[LINESIZE];
        char timeString[LINESIZE];

        qosAlarm = QOS_MIN_MESSAGE_TIME;

        (void)strcpy (timeString, ctime(&(qosdata[recno].timeStamp)));
        timeString[strlen(timeString)-1] = '\0';

        if( qosdata[recno].success == QOS_ABANDON_CHAR )
        {
            (void)sprintf(tmp,"%s%s%s%d%s%s", LASTATTEMPT, WASABANDONED, THEREHAVE, 
                       qosdata[recno].cnt, FAILURESABANDONS, timeString);
        }
        else
        { 
            (void)sprintf(tmp,"%s%s%s%d%s%s", LASTATTEMPT, FAILED, THEREHAVE, 
                       qosdata[recno].cnt, FAILURESABANDONS, timeString);
        } 
        if( qosdata[recno].avetime != -1 )
        {
            (void)sprintf(waitMessage, "%s%s%d%s.",tmp, WHENLAST, 
                           tookAbout(qosdata[recno].avetime), SECONDS);
        }
        else
        {
            (void)sprintf(waitMessage,"%s%s", tmp, NOTIMEYET);
        }
    }
    else if( qosdata[recno].avetime == -1 )
    {
        standardWaitMessage();
    }
    else
    {
        char tmp[LINESIZE];
 
        if( qosdata[recno].cnt == 1 )
        {
            (void)sprintf(tmp,"%s%d%s.",LAST, 
		tookAbout(qosdata[recno].avetime), SECONDS);
        }
        else if( qosdata[recno].cnt <= 4)
        {
            (void)sprintf(tmp,"%s%d%s.",LASTFEW, 
		tookAbout(qosdata[recno].avetime), SECONDS);
        }
        else
        {
        (void)sprintf(tmp,"%s%d%s.",USUALLYTAKES, 
		tookAbout(qosdata[recno].avetime), SECONDS); 
        }
        if( qosdata[recno].avetime <= 35 )
        {
            (void)sprintf(waitMessage,"%s%s", tmp, OVERTIME);
            qosAlarm = qosdata[recno].avetime + QOS_SEARCH_RETURN_MARGIN;
        }
        else
        {
            (void)sprintf(waitMessage,"%s%s", tmp, PLEASEWAITUNTIL);
            qosAlarm = QOS_MIN_MESSAGE_TIME;
        }
    }
    (void)strcat(waitMessage, CONTROLC);
}

#include <math.h>

static int
tookAbout(secs)
float secs;
{
int i;

        i = rint((double)secs);
        if (i > 10)
                i -= i % 5;
        return i;
}

static void
strip_ou(dest_dnstr, src_dnstr)
char *dest_dnstr;
char *src_dnstr;
{
    (void)strcpy(dest_dnstr, src_dnstr);

    while( *dest_dnstr != '\0' )
    {
        if(*dest_dnstr++ =='@')
        {
            if(!strncmp(dest_dnstr, "ou=", 3))
            {
            *(dest_dnstr-1) = '\0';
            break;
            }
        }
    }
}

void
drawstar()
{
    static char ani[] = { '|','/','-','\\','\0'};
    static char *cp=ani;

    (void)printf("%c\b",*cp);

    fflush(stdout);
 
    if(*++cp == '\0')
        cp = ani;
}

#else
deqos_stub()
{
}
#endif QOS
